﻿using System.Drawing.Drawing2D;

namespace KPGBase.Excercises
{
    internal class Excercise06 : BasicExcercise
    {
        public string lennaPath = "Lenna.png";

        // Matrices for image transformations
        private Matrix transformation;
        private Matrix rotation = new Matrix();
        private Matrix translation = new Matrix();

        // Mouse movement
        Point startPoint = new Point(0, 0);
        Point translationVector = new Point(0, 0);
        Point lastTranslationVector = new Point(0, 0);
        bool leftClicked;
        bool rightClicked;

        public Excercise06(Bitmap canvas) : base(canvas)
        {
            LoadImage(lennaPath);
            transformation = new Matrix(); // inits as identity matrix
        }

        /// <summary>
        /// Load image from path and set it as local canvas + create graphics
        /// </summary>
        /// <param name="path"> Path to image </param>
        public void LoadImage(string path)
        {
            Bitmap? lenna = LoadOriginalImg(path);
            if (lenna != null)
            {
                canvas = lenna;
                g = Graphics.FromImage(canvas);
            }
        }

        /// <summary>
        /// Load image from path
        /// </summary>
        /// <param name="path"> Path to image </param>
        /// <returns> Bitmap with loaded image </returns>
        private Bitmap? LoadOriginalImg(string path)
        {
            if (File.Exists(path))
                return (Bitmap)Image.FromFile(path);

            return null;
        }

        /// <summary>
        /// Converts image to black and white
        /// </summary>
        /// <returns> Edited image </returns>
        internal Bitmap BlackAndWhite()
        {
            // clone of the current bitmap - in it will be created the edited image created
            Bitmap bmp = (Bitmap)canvas.Clone();

            // Method that converts RGB image to greyscale image
            // Any grey color has all RGB components set to the same value
            // To obtain different grey tones through the image the values must correspond to the original pixel brightness
            // 
            // One pixel can be obtained by Bitmap.GetPixel(x,y) and then set as Bitmap.SetPixel(x, y, color)
            // Custom color can be created using Color.FromArgb(r, g, b)

            for (int i = 0; i < canvas.Width; i++)
            {
                for (int j = 0; j < canvas.Height; j++)
                {
                    Color c = canvas.GetPixel(i, j);
                    int gray = (int)((c.R + c.G + c.B) / 3.0);
                    // gray = (int)(0.299 * c.R + 0.587 * c.G + 0.114 * c.B);
                    bmp.SetPixel(i, j, Color.FromArgb(gray, gray, gray));
                }
            }

            canvas = bmp;
            Bitmap tfBmp = DrawTransformedBitmap();
            return tfBmp;
        }

        /// <summary>
        /// Applies embossing effect on the image
        /// </summary>
        /// <returns> Edited image </returns>
        internal Bitmap Emboss()
        {
            // clone of the current bitmap - in it will be created the edited image created
            Bitmap bmp = (Bitmap)canvas.Clone();

            // Method that applies embossing effect onto the image
            // Two pixels p1 = (i, j) and diagonal neighbor p2 = (i+1, j+1)
            // Get the difference of the values of every color channel separately p1 - p2
            // Add half of the allowed value interval to this difference
            // Set these values as new color

            Color pixel, neigh;
            for (int i = 0; i < canvas.Width - 1; i++)
            {
                for (int j = 0; j < canvas.Height - 1; j++)
                {
                    pixel = canvas.GetPixel(i, j);
                    neigh = canvas.GetPixel(i + 1, j + 1);

                    int red = Math.Min(Math.Max(pixel.R - neigh.R + 128, 0), 255);
                    int green = Math.Min(Math.Max(pixel.G - neigh.G + 128, 0), 255);
                    int blue = Math.Min(Math.Max(pixel.B - neigh.B + 128, 0), 255);

                    bmp.SetPixel(i, j, Color.FromArgb(red, green, blue));
                }
            }

            canvas = bmp;
            Bitmap tfBmp = DrawTransformedBitmap();
            return tfBmp;
        }

        /// <summary>
        /// Detects edges in the image
        /// </summary>
        /// <returns> Edited image </returns>
        internal Bitmap EdgeDetection()
        {
            // clone of the current bitmap - in it will be created the edited image created
            Bitmap bmp = (Bitmap)canvas.Clone();

            // Method that edge detection effect onto the image
            // For this we need four neighboring pixels
            // 1 2
            // 3 4
            // Compute the difference of pixel colours at the diagonals (eg. pixel 1 and 4, and pixel 2 and 3)
            // Add the differences and set them as the new color of pixel 1

            for (int i = 0; i < canvas.Width - 1; i++)
            {
                for (int j = 0; j < canvas.Height - 1; j++)
                {
                    Color p1 = canvas.GetPixel(i, j);
                    Color p2 = canvas.GetPixel(i + 1, j + 1);
                    Color p3 = canvas.GetPixel(i, j + 1);
                    Color p4 = canvas.GetPixel(i + 1, j);

                    int red = Math.Abs((int)p1.R - (int)p2.R)
                             + Math.Abs((int)p3.R - (int)p4.R);
                    int green = Math.Abs((int)p1.G - (int)p2.G)
                               + Math.Abs((int)p3.G - (int)p4.G);
                    int blue = Math.Abs((int)p1.B - (int)p2.B) 
                              + Math.Abs((int)p3.B - (int)p4.B);

                    red = Math.Min(red, 255);
                    green = Math.Min(green, 255);
                    blue = Math.Min(blue, 255);
                    int grey = (red + green + blue) / 3;

                    Color c = Color.FromArgb(grey, grey, grey); // (red, green, blue);
                    bmp.SetPixel(i, j, c);
                }
            }

            canvas = bmp;
            Bitmap tfBmp = DrawTransformedBitmap();
            return tfBmp;
        }

        /// <summary>
        /// Custom effect
        /// </summary>
        internal Bitmap OwnEffect()
        {
            Bitmap bmp = (Bitmap)canvas.Clone();

            // Space to create your own effect

            for (int i = 0; i < bmp.Width; i++)
            {
                for (int j = 0; j < bmp.Height; j++)
                {
                    Color pixel = bmp.GetPixel(i, j);

                    Color leftUpper = Color.FromArgb(0, 0, 0);
                    if (i > 0 && j > 0)
                    {
                        leftUpper = bmp.GetPixel(i - 1, j - 1);
                    }

                    int red = Math.Abs(pixel.R - leftUpper.R);
                    int green = Math.Abs(pixel.G - leftUpper.G);
                    int blue = Math.Abs(pixel.B - leftUpper.B);

                    int diff = red;
                    if (Math.Abs(green) > Math.Abs(diff))
                        diff = green;
                    if (Math.Abs(blue) > Math.Abs(diff))
                        diff = blue;

                    int gray = Math.Max(Math.Min(65 + diff, 255), 0);
                    bmp.SetPixel(i, j, Color.FromArgb(gray, gray, gray));
                }
            }

            canvas = bmp;
            Bitmap tfBmp = DrawTransformedBitmap();
            return tfBmp;
        }

        /// <summary>
        /// Mouse clicked
        /// </summary>
        internal void MouseClicked(Point location, MouseButtons button)
        {
            startPoint = location;

            // left mouse button
            if (button == MouseButtons.Left)
                leftClicked = true;
            // right mouse button
            else if (button == MouseButtons.Right)
                rightClicked = true;
        }

        /// <summary>
        /// Mose un-clicked
        /// </summary>
        internal void MouseUp()
        {
            leftClicked = false;
            rightClicked = false;
        }

        /// <summary>
        /// Mouse moved
        /// - if left clicked -> translate
        /// - if right clicked -> rotate
        /// </summary>
        /// <returns> Transformed image </returns>
        internal Bitmap MouseMove(Point newLoc)
        {
            // by how much did the mouse move
            translationVector.X = newLoc.X - startPoint.X;
            translationVector.Y = newLoc.Y - startPoint.Y;
            startPoint = newLoc;

            if (leftClicked)
            {
                // Translating image on the canvas
                
                /*
                 * 1) Urcit si vektor posunuti "translationVector" od predchozi pozice "startPoint"
                 * 2) Nastavte matici posunuti (translation.Translate(...))
                 *      - matici je vzdy nutne resetovat
                 * 3) Tuto matici podhodte metode drawTransformedBitmap
                 * 4) Aktualizujte pozice a vektory
                 * 
                 * !!! Metoda drawTransformedBitmap funguje tak, ze posouva vzdy puvodni obrazek !!!
                 *      - Je tedy nutne si nekam akumulovat vektor posunuti - napr. lastTranslationVector
                 */

                // save translation
                translation.Reset();
                translation.Translate(translationVector.X + lastTranslationVector.X,
                                      translationVector.Y + lastTranslationVector.Y);

                // save last translation vector
                lastTranslationVector.X += translationVector.X;
                lastTranslationVector.Y += translationVector.Y;

            }
            else if (rightClicked)
            {
                // save rotation
                rotation.Rotate(translationVector.X);
            }

            /*
                 * 1) Vytvorte si nejakou novou matici transformaci
                 *      - napr. Matrix transformation = new Matrix();
                 * 2) Vhodne transformujte obrazek
                 * 
                 * Pro rotaci lze pouzit metoda matice.Rotate(degree), 
                 * ktera orotuje obrazek kolem stredu soustavy souradnic 
                 * o uhel "degree"
                 * 
                 * Jako uhel muzete zvolit napr. vektor posunuti mysi 
                 * ve smeru osy x.
            */

            Matrix transformation = new Matrix();
            
            // add translation
            transformation.Multiply(translation);

            // move to the center of image
            Matrix toCenter = new Matrix();
            toCenter.Translate((float)canvas.Width / 2, (float)canvas.Height / 2);
            transformation.Multiply(toCenter);

            // rotate
            transformation.Multiply(rotation);

            // move back to 0,0
            Matrix fromCenter = new Matrix();
            fromCenter.Translate(-(float)canvas.Width / 2, -(float)canvas.Height / 2);
            transformation.Multiply(fromCenter);

            // draw the image
            this.transformation = transformation;
            Bitmap translated = DrawTransformedBitmap();
            return translated;
        }

        /// <summary>
        /// Draw bitmap transformed by transformation matrix
        /// </summary>
        /// <returns> Canvas with transformed image </returns>
        private Bitmap DrawTransformedBitmap()
        {
            Bitmap currImg = (Bitmap)canvas.Clone();

            //graphics access
            Graphics g = Graphics.FromImage(currImg);

            //clear canvas, apply transformations and draw image
            g.Clear(Color.White);
            g.Transform.Reset();
            g.Transform = transformation;

            g.DrawImage(canvas, 0, 0);
            return currImg;
        }
    }
}
