﻿using System.Drawing.Drawing2D;
using System.Drawing.Imaging;

namespace KPGBase.Excercises
{
    public class Particle
    {
        public PointF startposition; // Starting position
        public PointF position; // Current position
        public double rotation; // Current rotation
        public double rotationIncrement; // Rotation speed (angle increment each step)

        public float size;      // Particle size
        public PointF center;   // Center of the particle texture
        public int bitmapIndex; // Index of the used texture in texture array
        
        public int life;        // How old is the particle
        public int lifeEnd;     // Life expectancy of the particle
    }

    internal class Excercise10 : BasicExcercise
    {
        private Random random;
        
        // Textures
        private Bitmap bg;
        private Bitmap[] cloudTextures; // Availible particle textures

        private PointF coordOrigin;

        private List<Particle> particles;
        private int maxParticles;   

        // Min life length of a particle, max life length is lifeLength*3
        private int lifeLength;
        
        // Wind speed in y coordinate
        private float windSpeed; 
        
        // Increment of particle size
        private float sizeIncremnet;

        public Excercise10(Bitmap canvas) : base(canvas)
        {
            particles = new List<Particle>();
            random = new Random();

            // Set simulation parameters
            this.coordOrigin = new PointF(0.5f * canvas.Width, 0.86f * canvas.Height);
            this.maxParticles = 100;
            this.lifeLength = 80;
            this.windSpeed = 2f;
            this.sizeIncremnet = 0.01f;

            bg = (Bitmap)Image.FromFile("Data\\bg.png");

            cloudTextures = new Bitmap[3];
            cloudTextures[0] = (Bitmap)Image.FromFile("Data\\cloud_01.png");
            cloudTextures[1] = (Bitmap)Image.FromFile("Data\\cloud_02.png");
            cloudTextures[2] = (Bitmap)Image.FromFile("Data\\cloud_03.png");
        }

        /// <summary>
        /// Reset simulation
        /// </summary>
        public void ResetSimulation()
        {
            particles.Clear();
        }

        /// <summary>
        /// One simulation step
        /// </summary>
        public void RunSimulation()
        {
            GenerateParticle(maxParticles);
            DrawParticles(particles);     
        }

        /// <summary>
        /// Create a new particle if it is possible
        /// </summary>
        /// <param name="numOfParticles"> Max number of particles </param>
        private void GenerateParticle(int numOfParticles)
        {
            // if there is less particles than limit
            if (particles.Count < numOfParticles)
            {
                // generate one new particle and select a random texture
                Particle p = new Particle();
                p.bitmapIndex = random.Next(cloudTextures.Length);
                p.center = new PointF(cloudTextures[p.bitmapIndex].Width / 2,
                                      cloudTextures[p.bitmapIndex].Height / 2); // Center of the texture

                // generate starting position
                p.position = new PointF(random.Next(-1, 1), random.Next(-5, 5)); 
                p.startposition = p.position;

                // rotation
                p.rotation = random.Next(360);  
                p.rotationIncrement = (random.NextDouble() - 0.5) * 5;
                
                // particle size -> 0.05 - 0.1
                p.size = (float)random.NextDouble() * 0.05f + 0.05f;
                
                // particle life expectancy
                p.life = 0;                                       
                p.lifeEnd = random.Next(lifeLength, lifeLength * 3);

                // add to the list
                particles.Add(p); 
            }
        }

        // Vykresli celou mnozinu vygenerovanych castic
        private void DrawParticles(List<Particle> particles)
        {
            ClearImage();
            g.ResetTransform();
            g.DrawImage((Image)bg, new Rectangle(0, 0, canvas.Width, canvas.Height));

            // Draw all particles in the particle list
            // Simple option: (no rotation and no textures)
            // - compute transparency
            //      - the older the particle is the more transparent it is (= value of alpha is closer to 1.0)
            //      - at the end of the particle's life it should be fully transparent
            //      - value of transparency can change linearly or quadratically
            // - position of the particle
            //      - relative position is in the attribute of the particle class
            //      - don't forget the position of the origin (coordOrigin)
            // - fill rectangle/circle with computed transparency
            //      - size of the rectangle depends on the size of the particle (multiply it by 100 for better visibility)
            // Prettier option: (with textures and rotation)
            // - compute transparency
            // - transformation matrix: Matrix tmat
            //      - translation: translate to particle's position in the window
            //          - tmat.Multiply(translation)
            //      - scale: scale by particle's size
            //          - tmat.Multiply(scale)
            //      - rotation: rotate by particle's rotation
            //          - tmat.Multiply(rotation)
            //      - apply to graphics context: g.Transform = tmat;
            // - draw texture using the DrawAlphaMask method
            //      - choose the right texture for the specific particle
            //      - don't forget to apply the transparency

            // Update all of the particles' properties such as 
            // - life: life increases
            // - size: increases by sizeIncrement
            // - position: position is affected by the wind (attribute windSpeed) upwards
            // - rotation: increases by particle's rotationIncrement

            // Delete dead particles
            // - if particle's life is larger than it's life expectancy, remove it from the particle list

            // Transformacni matice pro jednu castici
            Matrix res = new Matrix();         // Matice vsech transformaci
            Matrix scale = new Matrix();       // Matice pro zmenu velikosti objektu
            Matrix translation = new Matrix(); // Matice translace (posunu)
            Matrix rotation = new Matrix();    // Matice rotace

            foreach (Particle p in particles) // Pro vsechny castice
            {
                // Jen čtverec

                // Pruhlednost castice v intervalu <0,1>, kde 1 je maximalni pruhlednost objektu
                float transparency = (float)Math.Min((double)(p.lifeEnd - p.life + 0.0001) / (float)p.lifeEnd, 1.0);
                transparency = (float)Math.Pow(transparency, 2);

                /*
                g.FillRectangle(new SolidBrush(Color.FromArgb((int)(255*transparency), 255, 255, 255)), 
                                new Rectangle((int)(coordOrigin.X + p.position.X - p.size * p.center.X),
                                              (int)(coordOrigin.Y + p.position.Y - p.size * p.center.Y),
                                              (int)(100*p.size), (int)(100*p.size)));
                */
                // Nebo nejdřív TF souřadného systému a vykreslit bitmapu

                // Resetovani vsech transformacnich matic
                scale.Reset();      // Resetuje matici zmeny velikosti objektu
                rotation.Reset();   // Resetuje matici rotace
                translation.Reset();// Resetuje matici translace (posuvu)

                // Spocitat transformacni matice
                scale.Scale(p.size, p.size);                                         // Matice zmeny velikosti meritka dane castice
                rotation.RotateAt((float)p.rotation, p.center);                      // Matice rotace dane castice
                translation.Translate(coordOrigin.X + p.position.X - p.size * p.center.X, // Matice posunuti
                                      coordOrigin.Y + p.position.Y - p.size * p.center.Y);

                // Vysledna transformacni matice
                res.Reset();                     // Resetovani matice transformaci
                res.Multiply(translation);       // Posune castici od stredu souradneho systemu o dany vektor
                res.Multiply(scale);             // Zvetsi nebo zmensi castici
                res.Multiply(rotation);          // Orotuje castici texturu castice 
                g.Transform = res;               // Nastavi transformacni matici na matici slozenych transformaci

                DrawAlphaMask(cloudTextures[p.bitmapIndex], 0, 0, 0, transparency); // Vykresli texturu castice s nastavenou pruhlednosti

                // Update all of the properties
                p.size += sizeIncremnet;// Zvetsi velikost castice
                p.rotation += p.rotationIncrement; // Zvysi uhel natoceni dane castice
                p.life++;                                     // Zvyseni casu existence castice

                p.position.Y -= windSpeed;
            }

            // Delete dead particles
            for (int i = 0; i < particles.Count; i++)         // Pro vsechny castice
            {
                if (particles[i].life > particles[i].lifeEnd) // Kdyz existuji dele nez by mely
                {
                    particles.RemoveAt(i);                    // Smaze castici
                    i--;
                }
            }
        }

        /// <summary>
        /// Draw png image with specific color overlay
        /// </summary>
        /// <param name="bitmap"> Bitmap to draw </param>
        /// <param name="redIntensity"> Red "filter" </param>
        /// <param name="greenIntensity"> Green "filter" </param>
        /// <param name="blueIntensity"> Blue "filter" </param>
        /// <param name="alphaIntensity"> Alpha of the drawn image </param>
        private void DrawAlphaMask(Bitmap bitmap, float redIntensity, float greenIntensity, float blueIntensity, float alphaIntensity)
        {
            // Transform matrix for alpha and color
            ColorMatrix matrix = new ColorMatrix();
            matrix.Matrix00 = redIntensity;
            matrix.Matrix11 = greenIntensity;
            matrix.Matrix22 = blueIntensity;
            matrix.Matrix33 = alphaIntensity;

            // Set color matrix
            ImageAttributes attr = new ImageAttributes();
            attr.SetColorMatrix(matrix, ColorMatrixFlag.Default, ColorAdjustType.Default);

            // Draw bitmap
            g.DrawImage((Image)bitmap,
                            new Rectangle(0, 0, bitmap.Width, bitmap.Height),
                            0, 0, bitmap.Width, bitmap.Height,
                            GraphicsUnit.Pixel, attr);
        }

        /// <summary>
        /// Set wind speed
        /// </summary>
        internal float WindSpeed
        {
            get { return this.windSpeed; }
            set { this.windSpeed = value; }
        }

        /// <summary>
        /// Set particle life
        /// </summary>
        internal int ParticleLifetime
        {
            get { return this.lifeLength; }
            set { this.lifeLength = value; }
        }

        /// <summary>
        /// Set particle size
        /// </summary>
        internal float ParticleSize
        {
            get { return this.sizeIncremnet; }
            set { this.sizeIncremnet = value; }
        }

    }
}
